package org.cyrano.timetable.ui.main;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Locale;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.cyrano.gimp.palette.GimpPalette;
import org.cyrano.gimp.palette.SequentialColorMap;
import org.cyrano.timetable.ui.edit.DisplayModeProvider;
import org.cyrano.timetable.ui.edit.prof.ProfTreePanel;
import org.cyrano.timetable.ui.edit.room.RoomTreePanel;
import org.cyrano.timetable.ui.edit.sems.SemsTreePanel;
import org.cyrano.timetable.ui.model.hour.HourModel;
import org.cyrano.timetable.ui.model.timetable.Root;
import org.cyrano.timetable.ui.reports.SemsReport;
import org.cyrano.timetable.ui.tool.AddTool;
import org.cyrano.timetable.ui.tool.PointerTool;
import org.cyrano.timetable.ui.view.MainPanel;
import org.cyrano.timetable.ui.view.TimeTable2XML;
import org.cyrano.timetable.ui.view.XML2TimeTable;
import org.cyrano.utils.ExtFileFilter;
import org.cyrano.utils.ImageCache;

/**
 * @author Demián Gutierrez
 */
public class MainFrame extends JFrame {

  private static final String[] ROWS = new String[]{ //
  "07", "08", "09", "10", "11", "12", "13", "14", "15", "16", "17", "18"};

  private static final String[] COLS = new String[]{ //
  "l", "m", "i", "j", "v"};

  private static final String GIMP_PALETTE = "resources/Custom.gpl";

  private static final int FRAME_W = 800;
  private static final int FRAME_H = 600;

  // --------------------------------------------------------------------------------

  public static JFrame workbenchRef;

  // --------------------------------------------------------------------------------

  private SemsTreePanel semsTreePanel;
  private ProfTreePanel profTreePanel;
  private RoomTreePanel roomTreePanel;

  private MainPanel mainPanel;

  private File currOpenFile;

  // --------------------------------------------------------------------------------

  //  private static final String TEST_TIMETABLE = "demo_timetable.xml";

  public MainFrame() {
    initWorkBenchRef();

    Root root = new Root(new HourModel(COLS, ROWS));

    //    try {
    //      root = loadRoot(new File(TEST_TIMETABLE));
    //    } catch (Exception e) {
    //      throw new RuntimeException(e);
    //    }

    Locale.setDefault(new Locale("es"));

    setLayout(new BorderLayout());

    // ----------------------------------------

    initMainPanel();
    mainPanel.setRoot(root);

    JTabbedPane treePanel = initTreePanel(root);

    JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
    splitPane.add(treePanel);
    splitPane.add(mainPanel);
    splitPane.setDividerLocation(300);
    add(splitPane, BorderLayout.CENTER);

    // ----------------------------------------

    add(initToolBarPanel(), BorderLayout.NORTH);

    // ----------------------------------------

    setDefaultCloseOperation(EXIT_ON_CLOSE);
    setSize(FRAME_W, FRAME_H);
    setVisible(true);
  }

  // --------------------------------------------------------------------------------

  private GimpPalette loadGimpPalette() {
    GimpPalette gimpPalette = new GimpPalette();

    try {
      gimpPalette.read( //
          ClassLoader.getSystemResourceAsStream(GIMP_PALETTE));
    } catch (Exception e) {
      throw new RuntimeException(e);
    }

    return gimpPalette;
  }

  // --------------------------------------------------------------------------------

  private void initWorkBenchRef() {
    if (workbenchRef != null) {
      throw new IllegalStateException();
    }

    workbenchRef = this;
  }

  // --------------------------------------------------------------------------------

  private JPanel initMainPanel() {
    if (mainPanel != null) {
      throw new IllegalStateException();
    }

    mainPanel = new MainPanel();

    JPanel ret = new JPanel(new BorderLayout());
    ret.setBorder(BorderFactory.createMatteBorder(//
        5, 5, 5, 5, Color.DARK_GRAY));
    ret.add(mainPanel, BorderLayout.CENTER);

    return ret;
  }

  // --------------------------------------------------------------------------------

  private JTabbedPane initTreePanel(Root timeTable) {

    JTabbedPane ret = new JTabbedPane();

    ret.addChangeListener(new ChangeListener() {
      @Override
      public void stateChanged(ChangeEvent evt) {
        treeTabbedPaneSelected(evt);
      }
    });

    semsTreePanel = new SemsTreePanel();
    semsTreePanel.setRoot(timeTable);
    ret.add(_I18NMainFrame.tabTtbl(), semsTreePanel);

    profTreePanel = new ProfTreePanel();
    profTreePanel.setRoot(timeTable);
    ret.add(_I18NMainFrame.tabProf(), profTreePanel);

    roomTreePanel = new RoomTreePanel();
    roomTreePanel.setRoot(timeTable);
    ret.add(_I18NMainFrame.tabRoom(), roomTreePanel);

    return ret;
  }

  // --------------------------------------------------------------------------------

  private void treeTabbedPaneSelected(ChangeEvent evt) {
    JTabbedPane tabbedPane = (JTabbedPane) evt.getSource();

    Component component = tabbedPane.getSelectedComponent();

    if (component instanceof DisplayModeProvider) {
      DisplayModeProvider displayModeProvider = //
      (DisplayModeProvider) component;
      mainPanel.setDisplayMode(displayModeProvider.getDisplayMode());
      displayModeProvider.updateSelectedItems();
    }
  }

  // --------------------------------------------------------------------------------

  private JComponent initToolBarPanel() {
    ButtonGroup buttonGroup = new ButtonGroup();

    JToolBar toolBar = new JToolBar();

    // --------------------------------------------------------------------------------

    JButton btnNew = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_new.png")));
    btnNew.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        btnNewClicked();
      }
    });
    toolBar.add(btnNew);

    // --------------------------------------------------------------------------------

    JButton btnSave = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_save.png")));
    btnSave.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        if (currOpenFile == null) {
          btnSaveAsClicked();
        } else {
          btnSaveClicked();
        }
      }
    });
    toolBar.add(btnSave);

    // --------------------------------------------------------------------------------

    JButton btnSaveAs = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_save_as.png")));
    btnSaveAs.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        btnSaveAsClicked();
      }
    });
    toolBar.add(btnSaveAs);

    // --------------------------------------------------------------------------------

    JButton btnOpen = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_open.png")));
    btnOpen.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        btnOpenClicked();
      }
    });
    toolBar.add(btnOpen);

    // --------------------------------------------------------------------------------

    toolBar.addSeparator();

    // --------------------------------------------------------------------------------

    JButton btnSpreadsheet = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_spreadsheet.png")));
    btnSpreadsheet.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        btnSemsReportClicked();
      }
    });
    toolBar.add(btnSpreadsheet);

    // --------------------------------------------------------------------------------

    JButton btnPicture = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_picture.png")));
    btnPicture.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        btnPictureClicked();
      }
    });
    toolBar.add(btnPicture);

    // --------------------------------------------------------------------------------

    toolBar.addSeparator();

    // --------------------------------------------------------------------------------

    JButton btnUndo = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_undo.png")));
    btnUndo.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        mainPanel.undo();
      }
    });
    toolBar.add(btnUndo);

    // --------------------------------------------------------------------------------

    JButton btnRedo = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_redo.png")));
    btnRedo.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        mainPanel.redo();
      }
    });
    toolBar.add(btnRedo);

    // --------------------------------------------------------------------------------

    toolBar.addSeparator();

    // --------------------------------------------------------------------------------

    JToggleButton btnSelect = new JToggleButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_cursor.png")));
    btnSelect.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        mainPanel.setCurrentTool(new PointerTool());
      }
    });
    toolBar.add(btnSelect);
    buttonGroup.add(btnSelect);

    // --------------------------------------------------------------------------------

    JToggleButton btnCreate = new JToggleButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_create.png")));
    btnCreate.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        mainPanel.setCurrentTool(new AddTool());
      }
    });
    toolBar.add(btnCreate);
    buttonGroup.add(btnCreate);

    // --------------------------------------------------------------------------------

    //    JToggleButton btnDelete = new JToggleButton( //
    //        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_delete.png")));
    //    btnDelete.addActionListener(new ActionListener() {
    //      public void actionPerformed(ActionEvent evt) {
    //        mainPanel.setCurrentTool(new DelTool());
    //      }
    //    });
    //    toolBar.add(btnDelete);
    //    buttonGroup.add(btnDelete);

    // --------------------------------------------------------------------------------

    toolBar.addSeparator();

    // --------------------------------------------------------------------------------

    JButton btnAbout = new JButton( //
        new ImageIcon(ImageCache.getInstance().getImage("images/24x24_about.png")));
    btnAbout.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        btnAboutClicked();
      }
    });
    toolBar.add(btnAbout);

    // --------------------------------------------------------------------------------

    btnSelect.doClick();

    return toolBar;
  }

  // --------------------------------------------------------------------------------

  private Root loadRoot(File file) throws Exception {
    XML2TimeTable xml2TimeTable = new XML2TimeTable();

    GimpPalette gimpPalette = loadGimpPalette();

    SequentialColorMap semsCmap = //
    new SequentialColorMap(gimpPalette);
    SequentialColorMap profCmap = //
    new SequentialColorMap(gimpPalette);
    SequentialColorMap roomCmap = //
    new SequentialColorMap(gimpPalette);

    return xml2TimeTable.load(new FileInputStream(file), //
        new HourModel(COLS, ROWS), semsCmap, profCmap, roomCmap);
  }

  // --------------------------------------------------------------------------------

  private void saveRoot(File file, Root root) throws Exception {
    // TODO: validations, empty root, etc
    TimeTable2XML timeTable2xml = new TimeTable2XML();

    timeTable2xml.save( //
        new FileOutputStream(file), mainPanel.getRoot());
  }

  // --------------------------------------------------------------------------------

  private JFileChooser createFileChooser(String ext) {
    JFileChooser fileChooser = new JFileChooser(currOpenFile);
    fileChooser.setAcceptAllFileFilterUsed(false);
    fileChooser.setMultiSelectionEnabled(false);
    fileChooser.setFileFilter(new ExtFileFilter(ext));

    return fileChooser;
  }

  // --------------------------------------------------------------------------------

  private File getFileChooserFile(JFileChooser fileChooser) {
    File file = fileChooser.getSelectedFile();
    String path = file.getAbsolutePath();

    ExtFileFilter extFileFilter = //
    (ExtFileFilter) fileChooser.getFileFilter();

    String ext = extFileFilter.getExt();

    if (!path.endsWith(ext)) {
      file = new File(path + "." + ext);
    }

    return file;
  }

  // --------------------------------------------------------------------------------

  private void btnNewClicked() {
    Root root = new Root(new HourModel(COLS, ROWS));

    mainPanel.setRoot(root);

    semsTreePanel.setRoot(root);
    profTreePanel.setRoot(root);
    roomTreePanel.setRoot(root);

    currOpenFile = null;
  }

  // --------------------------------------------------------------------------------

  private void btnSaveClicked() {
    try {
      saveRoot(currOpenFile, mainPanel.getRoot());
    } catch (Exception e) {
      JOptionPane.showMessageDialog(MainFrame.workbenchRef, //
          _I18NMainFrame.errorSaveError(), //
          _I18NMainFrame.errorTitle(), //
          JOptionPane.ERROR_MESSAGE);
      e.printStackTrace();
    }
  }

  // --------------------------------------------------------------------------------

  private void btnSaveAsClicked() {
    JFileChooser fc = createFileChooser("xml");
    int retVal = fc.showSaveDialog(this);

    if (retVal != JFileChooser.APPROVE_OPTION) {
      return;
    }

    currOpenFile = getFileChooserFile(fc);

    try {
      saveRoot(currOpenFile, mainPanel.getRoot());
    } catch (Exception e) {
      JOptionPane.showMessageDialog(MainFrame.workbenchRef, //
          _I18NMainFrame.errorSaveError(), //
          _I18NMainFrame.errorTitle(), //
          JOptionPane.ERROR_MESSAGE);
      e.printStackTrace();
    }
  }

  // --------------------------------------------------------------------------------

  private void btnOpenClicked() {
    JFileChooser fc = createFileChooser("xml");
    int retVal = fc.showOpenDialog(this);

    if (retVal != JFileChooser.APPROVE_OPTION) {
      return;
    }

    currOpenFile = fc.getSelectedFile();

    Root root;

    try {
      root = loadRoot(currOpenFile);
    } catch (Exception e) {
      JOptionPane.showMessageDialog(MainFrame.workbenchRef, //
          _I18NMainFrame.errorOpenError(), //
          _I18NMainFrame.errorTitle(), //
          JOptionPane.ERROR_MESSAGE);
      e.printStackTrace();
      return;
    }

    mainPanel.setRoot(root);

    semsTreePanel.setRoot(root);
    profTreePanel.setRoot(root);
    roomTreePanel.setRoot(root);
  }

  // --------------------------------------------------------------------------------

  private void btnSemsReportClicked() {
    JFileChooser fc = createFileChooser("xls");
    int retVal = fc.showSaveDialog(this);

    if (retVal != JFileChooser.APPROVE_OPTION) {
      return;
    }

    File file = getFileChooserFile(fc);

    SemsReport report = new SemsReport(mainPanel.getRoot());
    report.generate();

    try {
      report.save(file);
    } catch (Exception e) {
      JOptionPane.showMessageDialog(MainFrame.workbenchRef, //
          _I18NMainFrame.errorSaveError(), //
          _I18NMainFrame.errorTitle(), //
          JOptionPane.ERROR_MESSAGE);
      e.printStackTrace();
    }
  }

  // --------------------------------------------------------------------------------

  private void btnPictureClicked() {
    JFileChooser fc = createFileChooser("png");
    int retVal = fc.showSaveDialog(this);

    if (retVal != JFileChooser.APPROVE_OPTION) {
      return;
    }

    int w = 1280;
    int h = 1024;

    BufferedImage bimg = new BufferedImage( //
        w, h, BufferedImage.TYPE_INT_RGB);
    Graphics2D g2d = (Graphics2D) bimg.getGraphics();

    mainPanel.update(g2d, w, h);

    g2d.dispose();

    File file = getFileChooserFile(fc);

    try {
      ImageIO.write(bimg, "png", file);
    } catch (Exception e) {
      JOptionPane.showMessageDialog(MainFrame.workbenchRef, //
          _I18NMainFrame.errorSaveError(), //
          _I18NMainFrame.errorTitle(), //
          JOptionPane.ERROR_MESSAGE);
      e.printStackTrace();
    }
  }

  // --------------------------------------------------------------------------------

  private void btnAboutClicked() {
    DlgAbout dlgAbout = new DlgAbout(workbenchRef);
    dlgAbout.setVisible(true);
  }

  // --------------------------------------------------------------------------------

  public static void main(String[] args) {
    new MainFrame();
  }
}
