package org.cyrano.timetable.ui.view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JPopupMenu;

import org.cyrano.editor.listeners.CommandEvent;
import org.cyrano.editor.listeners.CommandListener;
import org.cyrano.editor.listeners.SelectionEvent;
import org.cyrano.editor.listeners.SelectionListener;
import org.cyrano.editor.listeners.SelectionManager;
import org.cyrano.editor.listeners.UndoRedoManager;
import org.cyrano.editor.parts.ContainerViewPart;
import org.cyrano.editor.parts.EditPart;
import org.cyrano.editor.parts.PartsFactory;
import org.cyrano.editor.parts.PartsRegister;
import org.cyrano.editor.parts.ViewPart;
import org.cyrano.editor.tool.EditPolicyMouseHandler;
import org.cyrano.editor.tool.Tool;
import org.cyrano.timetable.ui.algorithm.LayoutAlgorithm;
import org.cyrano.timetable.ui.figures.TimeTableViewPart;
import org.cyrano.timetable.ui.model.base.BaseModel;
import org.cyrano.timetable.ui.model.base.StructureChangeEvent;
import org.cyrano.timetable.ui.model.base.StructureChangeListener;
import org.cyrano.timetable.ui.model.base.VisibilityChangeEvent;
import org.cyrano.timetable.ui.model.base.VisibilityChangeListener;
import org.cyrano.timetable.ui.model.timetable.DisplayMode;
import org.cyrano.timetable.ui.model.timetable.Root;
import org.cyrano.timetable.ui.model.timetable.Slot;
import org.cyrano.timetable.ui.model.visitor.ModelProcessor;
import org.cyrano.timetable.validator.RestrictionValidator;
import org.cyrano.timetable.validator.RestrictionValidatorFactory;
import org.cyrano.utils.Hwh;
import org.cyrano.utils.Rect;

public class MainPanel extends JPanel //
    implements
      CommandListener,
      VisibilityChangeListener,
      StructureChangeListener {

  private static final Stroke SELECTION_STROKE = //
  new BasicStroke(3, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 0, //
      new float[]{3, 10}, 0);

  private Root root;

  private EditPolicyMouseHandler editPolicy;

  private PartsRegister partsRegister;

  private Tool currentTool;

  // --------------------------------------------------------------------------------

  public MainPanel() {
    UndoRedoManager.getInstance().getCommandListenerProxy(). //
        addCommandListener(this);

    MouseAdapter mouseAdapter = new MouseAdapter() {
      @Override
      public void mouseReleased(MouseEvent evt) {
        MainPanel.this.mouseReleased(evt);
      }

      @Override
      public void mousePressed(MouseEvent evt) {
        MainPanel.this.mousePressed(evt);
      }

      @Override
      public void mouseDragged(MouseEvent evt) {
        MainPanel.this.mouseDragged(evt);
      }
    };

    addMouseMotionListener(mouseAdapter);
    addMouseListener(mouseAdapter);

    SelectionManager.getInstance().getSelectionListenerProxy(). //
        addSelectionListener(new SelectionListener() {
          @Override
          public void selectionChanged(SelectionEvent evt) {
            repaint();
          }
        });
  }

  // --------------------------------------------------------------------------------

  public Root getRoot() {
    return root;
  }

  // --------------------------------------------------------------------------------

  public void setRoot(Root timeTable) {
    if (this.root != null) {
      this.root.getVisibilityChangeSupport(). //
          delVisibilityChangeListener(this);
      this.root.getStructureChangeSupport(). //
          delStructureChangeListener(this);
    }

    this.root = timeTable;

    timeTable.getVisibilityChangeSupport(). //
        addVisibilityChangeListener(this);
    timeTable.getStructureChangeSupport(). //
        addStructureChangeListener(this);

    partsRegister = new PartsRegister(new PartsFactory());

    validateTimeTable();
    rebuildFigures();
    repaint();
  }

  // --------------------------------------------------------------------------------

  private void rebuildFigures() {
    PartsBuilderVisitor figureBuilderVisitor = //
    new PartsBuilderVisitor(partsRegister);

    ModelProcessor timeTableProcessor = //
    new ModelProcessor(figureBuilderVisitor);

    timeTableProcessor.visit(root);

    repaint();
  }

  // --------------------------------------------------------------------------------

  private void validateTimeTable() {
    // TODO: INJECT?
    RestrictionValidatorFactory restrictionValidatorFactory = //
    new RestrictionValidatorFactory();

    RestrictionValidator restrictionValidator = //
    restrictionValidatorFactory.createRestrictionValidator();

    try {
      restrictionValidator.validate(root);
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  // --------------------------------------------------------------------------------
  // Painting
  // --------------------------------------------------------------------------------

  @Override
  public void paint(Graphics g) {
    update(g);
  }

  // --------------------------------------------------------------------------------

  @Override
  public void update(Graphics g) {
    Graphics2D g2d = (Graphics2D) g;

    int w = Hwh.getW(this);
    int h = Hwh.getH(this);

    update(g2d, w, h);
  }

  // --------------------------------------------------------------------------------

  public void update(Graphics2D g2d, int w, int h) {
    g2d.setRenderingHint( //
        RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

    Rect rect = new Rect(0, 0, w, h);

    LayoutAlgorithm layoutAlgorithm = new LayoutAlgorithm(root);

    layoutAlgorithm.layout( //
        TimeTableViewPart.LEGEND_XOFFSET, //
        TimeTableViewPart.LEGEND_YOFFSET, //
        rect.w - TimeTableViewPart.LEGEND_XOFFSET, //
        rect.h - TimeTableViewPart.LEGEND_YOFFSET);

    ContainerViewPart containerViewPart = //
    (ContainerViewPart) partsRegister.getViewPartByModel(root);

    containerViewPart.setRect(rect);
    containerViewPart.paint(g2d);

    paintSelectedObjects(g2d, //
        SelectionManager.getInstance().getSelectedObjects());

    if (editPolicy != null) {
      editPolicy.paint(g2d);
    }
  }

  // --------------------------------------------------------------------------------

  public void paintSelectedObjects(Graphics2D g2d, List<Object> selectedObjects) {
    for (Object selectedObject : selectedObjects) {
      if (selectedObject instanceof Root) {
        // ----------------------------------------
        // Root shoud never be selected
        // ----------------------------------------
        continue;
      }

      List<ViewPart> viewPartList = //
      getViewPartsForSelectedObject(selectedObject);

      for (ViewPart viewPart : viewPartList) {
        Rect rect = viewPart.getRect();

        g2d.setColor(Color.WHITE);
        g2d.drawRect(rect.x, rect.y, rect.w, rect.h);

        Stroke prevStroke = g2d.getStroke();

        g2d.setStroke(SELECTION_STROKE);
        g2d.setColor(Color.BLACK);
        g2d.drawRect(rect.x, rect.y, rect.w, rect.h);

        g2d.setStroke(prevStroke);
      }
    }
  }

  // --------------------------------------------------------------------------------
  // MouseAdapter
  // --------------------------------------------------------------------------------

  private List<ViewPart> getViewPartsForSelectedObject(Object selectedObject) {
    BaseModel baseModel = (BaseModel) selectedObject;

    List<ViewPart> ret = new ArrayList<ViewPart>();

    for (Slot slot : baseModel.getAllChildrenSlots()) {
      ret.add(partsRegister.getViewPartByModel(slot));
    }

    return ret;
  }

  // --------------------------------------------------------------------------------

  protected void mouseReleased(MouseEvent evt) {
    if (editPolicy != null) {
      editPolicy.mouseClickUp(evt);
    }

    editPolicy = null;

    repaint();
  }

  // --------------------------------------------------------------------------------

  protected void mousePressed(MouseEvent evt) {
    if (currentTool == null) {
      return;
    }

    ContainerViewPart containerViewPart = //
    (ContainerViewPart) partsRegister.getViewPartByModel(root);

    // get view
    ViewPart viewPart = containerViewPart.getViewPartAt(evt.getPoint());

    // get edit
    EditPart editPart = partsRegister.getEditPartByViewPart(viewPart);

    SelectionManager.getInstance().clear();
    SelectionManager.getInstance().add( //
        partsRegister.getModelByEditPart(editPart));

    if (evt.isPopupTrigger()) {
      // popup
      JPopupMenu popupMenu = editPart.getPopupMenu(currentTool, evt);

      if (popupMenu != null) {
        popupMenu.show(this, evt.getPoint().x, evt.getPoint().y);
      }
    } else {
      // get policy (if any)
      editPolicy = editPart.getEditPolicy(currentTool, evt);
    }

    if (editPolicy != null) {
      editPolicy.mouseClickDw(evt);
    }

    repaint();
  }

  // --------------------------------------------------------------------------------

  protected void mouseDragged(MouseEvent evt) {
    if (editPolicy != null) {
      editPolicy.mouseDragged(evt);
    }

    repaint();
  }

  // --------------------------------------------------------------------------------
  // Tool / Undo / Redo Support
  // --------------------------------------------------------------------------------

  public void setDisplayMode(DisplayMode displayMode) {
    root.setDisplayMode(displayMode);
    rebuildFigures(); // TODO REQUIRED?
  }

  // --------------------------------------------------------------------------------

  public void setCurrentTool(Tool currentTool) {
    this.currentTool = currentTool;
  }

  // --------------------------------------------------------------------------------

  @Override
  public void commandFired(CommandEvent evt) {
    editPolicy = null;

    validateTimeTable();
    rebuildFigures(); // TODO REQUIRED?
    repaint();
  }

  // --------------------------------------------------------------------------------

  public void undo() {
    UndoRedoManager.getInstance().undo();
    validateTimeTable();
    repaint();
  }

  // --------------------------------------------------------------------------------

  public void redo() {
    UndoRedoManager.getInstance().redo();
    validateTimeTable();
    repaint();
  }

  // --------------------------------------------------------------------------------
  // VisibilityChangeListener
  // --------------------------------------------------------------------------------

  @Override
  public void visibilityChanged(VisibilityChangeEvent evt) {
    repaint();
  }

  // --------------------------------------------------------------------------------
  // StructureChangeListener
  // --------------------------------------------------------------------------------

  @Override
  public void addChild(StructureChangeEvent evt) {
    rebuildFigures(); // TODO REQUIRED?
  }

  @Override
  public void delChild(StructureChangeEvent evt) {
    SelectionManager.getInstance().clear();
    rebuildFigures(); // TODO REQUIRED?
  }

  @Override
  public void reparent(StructureChangeEvent evt) {
    validateTimeTable();
    rebuildFigures(); // TODO REQUIRED?
  }
}
